"""
Possible functions for the ``PRIVATE_STORAGE_AUTH_FUNCTION`` setting.
"""


def allow_authenticated(private_file):
    return private_file.request.user.is_authenticated()


def allow_staff(private_file):
    request = private_file.request
    return request.user.is_authenticated() and request.user.is_staff


def allow_superuser(private_file):
    request = private_file.request
    return request.user.is_authenticated() and request.user.is_superuser



"""
Django Storage interface
"""
from django.utils.module_loading import import_string

from private_storage import appconfig

__all__ = (
    'private_storage',
    'PrivateStorage',
)

# Fetch the storage class
PrivateStorage = import_string(appconfig.PRIVATE_STORAGE_CLASS)

# Singleton instance.
private_storage = PrivateStorage()



import mimetypes
from django.core.files.storage import Storage, File

from django.utils.functional import cached_property


class PrivateFile(object):
    """
    A wrapper object that describes the file that is being accessed.
    """

    def __init__(self, request, storage, relative_name):
        self.request = request
        self.storage = storage  # type: Storage
        self.relative_name = relative_name

    @cached_property
    def full_path(self):
        # Not using self.storage.open() as the X-Sendfile needs a normal path.
        return self.storage.path(self.relative_name)

    def open(self, mode='rb'):
        """
        Open the file for reading.
        :rtype: django.core.files.storage.File
        """
        file = self.storage.open(self.relative_name, mode=mode)  # type: File
        return file

    def exists(self):
        """
        Check whether the file exists.
        """
        return self.storage.exists(self.relative_name)

    @cached_property
    def content_type(self):
        """
        Return the HTTP ``Content-Type`` header value for a filename.
        """
        mimetype, encoding = mimetypes.guess_type(self.relative_name)
        return mimetype or 'application/octet-stream'

    @cached_property
    def size(self):
        """
        Return the size of the file in bytes.
        """
        return self.storage.size(self.relative_name)

    @cached_property
    def modified_time(self):
        """
        Return the last-modified time
        """
        return self.storage.get_modified_time(self.relative_name)



"""
Django Storage interface, using the file system backend.
"""
from django.core.files.storage import FileSystemStorage
from django.core.urlresolvers import reverse_lazy
from django.utils.deconstruct import deconstructible
from django.utils.encoding import force_text

from private_storage import appconfig


@deconstructible
class PrivateFileSystemStorage(FileSystemStorage):
    """
    Interface to the Django storage system,
    storing the files in a private folder.
    """
    def __init__(self, location=None, base_url=None, **kwargs):
        if location is None:
            location = appconfig.PRIVATE_STORAGE_ROOT

        super(PrivateFileSystemStorage, self).__init__(
            location=location,
            base_url=base_url,
            **kwargs
        )
        if base_url is None:
            # When base_url is not given, it's autodetected.
            # However, as the super method checks for base_url.endswith('/'),
            # the attribute is overwritten here to avoid breaking lazy evaluation.
            self.base_url = reverse_lazy('serve_private_file', kwargs={'path': ''})

    def url(self, name):
        # Make sure reverse_lazy() is evaluated, as Python 3 won't do this here.
        self.base_url = force_text(self.base_url)
        return super(PrivateFileSystemStorage, self).url(name)



"""
Sending files efficiently for different kind of webservers.
"""
import os

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import FileResponse, HttpResponse
from django.http import HttpResponseNotModified
from django.utils.http import http_date
from django.utils.lru_cache import lru_cache
from django.utils.module_loading import import_string
from django.views.static import serve, was_modified_since


@lru_cache()
def get_server_class(path):
    if '.' in path:
        return import_string(path)
    elif path == 'streaming':
        return DjangoStreamingServer
    elif path == 'django':
        return DjangoServer
    elif path == 'apache':
        return ApacheXSendfileServer
    elif path == 'nginx':
        return NginxXAccelRedirectServer
    else:
        raise ImproperlyConfigured(
            "PRIVATE_STORAGE_SERVER setting should be 'nginx', 'apache', 'django' or a python class path."
        )


class DjangoStreamingServer(object):
    """
    Serve static files as streaming chunks in Django.
    """

    @staticmethod
    def serve(private_file):
        # Support If-Last-Modified
        mtime = private_file.modified_time.timestamp()
        size = private_file.size
        if not was_modified_since(private_file.request.META.get('HTTP_IF_MODIFIED_SINCE'), mtime, size):
            return HttpResponseNotModified()

        # FileResponse will submit the file in 8KB chunks
        response = FileResponse(private_file.open())
        response['Content-Type'] = private_file.content_type
        response['Content-Length'] = size
        response["Last-Modified"] = http_date(mtime)
        return response


class DjangoServer(DjangoStreamingServer):
    """
    Serve static files from the local filesystem through Django.
    This is a bad idea for most situations other than testing.
    The file data is copied multiple times; read by Django, written to WSGI,
    read by webserver, outputted to client.
    """

    @staticmethod
    def serve(private_file):
        # This supports If-Modified-Since and sends the file in 8KB chunks
        try:
            full_path = private_file.full_path
        except NotImplementedError:
            # S3 files, fall back to streaming server
            return DjangoStreamingServer.serve(private_file)
        else:
            # Using Django's serve gives If-Modified-Since support out of the box.
            return serve(private_file.request, full_path, document_root='/', show_indexes=False)


class ApacheXSendfileServer(object):
    """
    Serve files for Apache with ``X-Sendfile``.
    """

    @staticmethod
    def serve(private_file):
        response = HttpResponse()
        response['X-Sendfile'] = private_file.full_path
        response['Content-Type'] = private_file.content_type
        return response


class NginxXAccelRedirectServer(object):
    """
    Serve the files for Nginx with ``X-Accel-Redirect``.
    Add the following configuration::

        location /private-x-accel-redirect/ (
            internal;
            alias /home/user/my/path/to/private/media/;
        )

    Or update the ``PRIVATE_STORAGE_INTERNAL_URL`` setting to use a different URL prefix.
    """

    @staticmethod
    def serve(private_file):
        internal_url = os.path.join(settings.PRIVATE_STORAGE_INTERNAL_URL, private_file.relative_name)
        response = HttpResponse()
        response['X-Accel-Redirect'] = internal_url
        response['Content-Type'] = private_file.content_type
        return response



"""
Views to send private files.
"""
from django.http import HttpResponseForbidden, Http404
from django.utils.module_loading import import_string
from django.views.generic import View
from django.views.generic.detail import SingleObjectMixin

from . import appconfig
from .models import PrivateFile
from .servers import get_server_class
from .storage import private_storage


class PrivateStorageView(View):
    """
    Return the uploaded files
    """

    #: The storage class to retrieve files from
    storage = private_storage

    #: The authorisation rule for accessing
    can_access_file = staticmethod(import_string(appconfig.PRIVATE_STORAGE_AUTH_FUNCTION))

    #: Import the server class once
    server_class = get_server_class(appconfig.PRIVATE_STORAGE_SERVER)

    def get_path(self):
        """
        Determine the path for the object to provide.
        This can be overwritten to combine the view with a different object retrieval.
        """
        return self.kwargs['path']

    def get_private_file(self):
        """
        Return all relevant data in a single object, so this is easy to extend
        and server implementations can pick what they need.
        """
        return PrivateFile(
            request=self.request,
            storage=self.storage,
            relative_name=self.get_path()
        )

    def get(self, request, *args, **kwargs):
        """
        Handle incoming GET requests
        """
        private_file = self.get_private_file()

        if not self.can_access_file(private_file):
            return HttpResponseForbidden('Private storage access denied')

        if not private_file.exists():
            raise Http404("File not found")

        return self.serve_file(private_file)

    def serve_file(self, private_file):
        """
        Serve the file that was retrieved from the storage.
        The relative path can be found with ``private_file.relative_name``.

        :type private_file: :class:`private_storage.models.PrivateFile`
        :rtype: django.http.HttpResponse
        """
        return self.server_class().serve(private_file)


class PrivateStorageDetailView(SingleObjectMixin, PrivateStorageView):
    """
    Download a document based on an object ID.
    This view can by used by third-party apps to implement their own download view.

    Implement access controls by overriding :meth`get_queryset` or redefining :meth:`can_access_file`.
    """
    model = None  #
    model_file_field = 'file'

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    def get_path(self):
        file = getattr(self.object, 'file')
        return file.name

    def can_access_file(self, private_file):
        """
        The authorization rule for this view.
        By default it reuses the ``PRIVATE_STORAGE_AUTH_FUNCTION`` setting,
        but this should likely be redefined.
        """
        return PrivateStorageView.can_access_file(private_file)



#-*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import os

import django
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import UploadedFile
from django.db import models
from django.template.defaultfilters import filesizeformat
from django.utils.encoding import smart_str
from django.utils.translation import ugettext_lazy as _
from .storage import private_storage

logger = logging.getLogger(__name__)



class PrivateFileField(models.FileField):
    """
    Filefield with private storage, custom filename and size checks.

    Extra settings:
    - ``upload_subfolder``: a lambda to find the subfolder, depending on the instance.
    - ``content_types``: list of allowed content types.
    - ``max_file_size``: maximum file size.
    """
    default_error_messages = {
        'invalid_file_type': _('File type not supported.'),
        'file_too_large': _('The file may not be larger then {max_size}.'),
    }

    def __init__(self, *args, **kwargs):
        self._upload = kwargs.pop('upload_subfolder', None)
        self.content_types = kwargs.pop("content_types", None) or ()
        self.max_file_size = kwargs.pop("max_file_size", None)

        kwargs.setdefault('storage', private_storage)
        if self._upload and django.VERSION < (1,7):
            kwargs.setdefault('upload_to', 'uploads')  # shut up warnings from Django 1.6- model validation
        super(PrivateFileField, self).__init__(*args, **kwargs)

    def clean(self, *args, **kwargs):
        data = super(PrivateFileField, self).clean(*args, **kwargs)
        file = data.file
        if isinstance(file, UploadedFile):
            # content_type is only available for uploaded files,
            # and not for files which are already stored in the model.
            content_type = file.content_type

            if self.content_types and content_type not in self.content_types:
                logger.debug('Rejected uploaded file type: %s', content_type)
                raise ValidationError(self.error_messages['invalid_file_type'])

            if self.max_file_size and file.size > self.max_file_size:
                raise ValidationError(self.error_messages['file_too_large'].format(
                    max_size=filesizeformat(self.max_file_size),
                    size=filesizeformat(file.size)
                ))

        return data

    def generate_filename(self, instance, filename):
        subdir_func = self._upload
        if subdir_func:
            subdirs = [self.storage.get_valid_name(dir) for dir in subdir_func(instance)]
        else:
            subdirs = [self.upload_to]

        if not subdirs:
            subdirs = [self.get_directory_name()]
        dirs = list(subdirs) + [self.get_filename(filename)]
        return smart_str(os.path.join(*dirs))



from django.conf import settings

PRIVATE_STORAGE_CLASS = getattr(settings, 'PRIVATE_STORAGE_CLASS', 'private_storage.storage.files.PrivateFileSystemStorage')
PRIVATE_STORAGE_ROOT = getattr(settings, 'PRIVATE_STORAGE_ROOT', None)
PRIVATE_STORAGE_SERVER = getattr(settings, 'PRIVATE_STORAGE_SERVER', 'django')
PRIVATE_STORAGE_AUTH_FUNCTION = getattr(settings, 'PRIVATE_STORAGE_AUTH_FUNCTION', 'private_storage.permissions.allow_superuser')

# For Nginx X-Accel-Redirect
PRIVATE_STORAGE_INTERNAL_URL = getattr(settings, 'PRIVATE_STORAGE_INTERNAL_URL', '/private-x-accel-redirect/')

PRIVATE_STORAGE_S3_REVERSE_PROXY = getattr(settings, 'PRIVATE_STORAGE_S3_REVERSE_PROXY', False)



from django.urls import reverse
from django.utils.deconstruct import deconstructible
from storages.backends.s3boto3 import S3Boto3Storage
from storages.utils import setting
from private_storage import appconfig


@deconstructible
class PrivateS3BotoStorage(S3Boto3Storage):
    """
    Private storage bucket for S3
    """
    # Since this class inherits the default storage, it shares many parameters with the base class.
    # Thus, redefine the setting name that is used to read these values, so almost all settings are not shared.
    access_key = setting('AWS_PRIVATE_S3_ACCESS_KEY_ID', setting('AWS_PRIVATE_ACCESS_KEY_ID', S3Boto3Storage.access_key))
    secret_key = setting('AWS_PRIVATE_S3_SECRET_ACCESS_KEY', setting('AWS_PRIVATE_SECRET_ACCESS_KEY', S3Boto3Storage.secret_key))
    file_overwrite = setting('AWS_PRIVATE_S3_FILE_OVERWRITE', False)  # false, differ from base class
    object_parameters = setting('AWS_PRIVATE_S3_OBJECT_PARAMETERS', {})
    bucket_name = setting('AWS_PRIVATE_STORAGE_BUCKET_NAME', strict=True)
    auto_create_bucket = setting('AWS_PRIVATE_AUTO_CREATE_BUCKET', False)
    default_acl = setting('AWS_PRIVATE_DEFAULT_ACL', 'private')  # differ from base class
    bucket_acl = setting('AWS_PRIVATE_BUCKET_ACL', default_acl)
    querystring_auth = setting('AWS_PRIVATE_QUERYSTRING_AUTH', True)
    querystring_expire = setting('AWS_PRIVATE_QUERYSTRING_EXPIRE', 3600)
    signature_version = setting('AWS_PRIVATE_S3_SIGNATURE_VERSION')
    reduced_redundancy = setting('AWS_PRIVATE_REDUCED_REDUNDANCY', False)
    location = setting('AWS_PRIVATE_LOCATION', '')
    encryption = setting('AWS_PRIVATE_S3_ENCRYPTION', False)
    custom_domain = setting('AWS_PRIVATE_S3_CUSTOM_DOMAIN')
    addressing_style = setting('AWS_PRIVATE_S3_ADDRESSING_STYLE')
    secure_urls = setting('AWS_PRIVATE_S3_SECURE_URLS', True)
    file_name_charset = setting('AWS_PRIVATE_S3_FILE_NAME_CHARSET', 'utf-8')
    gzip = setting('AWS_PRIVATE_IS_GZIPPED', S3Boto3Storage.gzip)  # fallback to default
    preload_metadata = setting('AWS_PRIVATE_PRELOAD_METADATA', False)
    url_protocol = setting('AWS_PRIVATE_S3_URL_PROTOCOL', S3Boto3Storage.url_protocol)  # fallback to default
    endpoint_url = setting('AWS_PRIVATE_S3_ENDPOINT_URL', None)
    region_name = setting('AWS_PRIVATE_S3_REGION_NAME', S3Boto3Storage.region_name)  # fallback to default
    use_ssl = setting('AWS_PRIVATE_S3_USE_SSL', True)

    def url(self, name, *args, **kwargs):
        if appconfig.PRIVATE_STORAGE_S3_REVERSE_PROXY or not self.querystring_auth:
            # There is no direct URL possible, return our streaming view instead.
            return reverse('serve_private_file', kwargs={'path': name})
        else:
            # The S3Boto3Storage can generate a presigned URL that is temporary available.
            return super(PrivateS3BotoStorage, self).url(name, *args, **kwargs)


@deconstructible
class PrivateEncryptedS3BotoStorage(PrivateS3BotoStorage):
    """
    Enforced encryption for private storage on S3.
    This is a convience option, it can also be implemented
    through :class:`PrivateS3BotoStorage` by using the proper settings.
    """
    encryption = True
    signature_version = PrivateS3BotoStorage.signature_version or 's3v4'
